@@ -1,11 +1,11 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2004 Nokia. All rights reserved.
+ * Copyright (C) 2004 Nokia Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * published by the Free Software Foundation; version 2 of the
+ * License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -26,9 +26,13 @@
 
 typedef struct _DaemonConnection DaemonConnection;
 
-DaemonConnection *daemon_connection_setup  (DBusConnection   *dbus_conn);
-void              daemon_connection_cancel (DaemonConnection *conn,
-					    gint32            cancellation_id);
+typedef void (*DaemonConnectionShutdownFunc) (DaemonConnection *conn,
+					      gpointer          data);
 
+DaemonConnection *daemon_connection_setup  (DBusConnection               *dbus_conn,
+					    DaemonConnectionShutdownFunc  shutdown_func,
+					    gpointer                      shutdown_data);
+void              daemon_connection_cancel (DaemonConnection             *conn,
+					    gint32                        cancellation_id);
 
 #endif /* __DAEMON_CONNECTION_H__ */
