@@ -1,11 +1,11 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2004 Nokia. All rights reserved.
+ * Copyright (C) 2004-2006 Nokia Corporation.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public License as
- * published by the Free Software Foundation; either version 2 of the
- * License, or (at your option) any later version.
+ * published by the Free Software Foundation; version 2 of the
+ * License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -26,6 +26,9 @@
 #include <libgnomevfs/gnome-vfs.h>
 #include <libgnomevfs/gnome-vfs-method.h>
 #include <libgnomevfs/gnome-vfs-dbus-utils.h>
+#include <libgnomevfs/gnome-vfs-module-shared.h>
+
+/*#include <pthread.h>*/
 
 #define DBUS_API_SUBJECT_TO_CHANGE
 #include <dbus/dbus-glib.h>
@@ -39,7 +42,7 @@
 
 #include "dbus-utils.h"
 
-#define d(x) 
+#define d(x)
 
 typedef struct {
 	gint32   handle_id;
@@ -54,32 +57,36 @@ typedef struct {
 	/* Probably need more stuff here? */
 } FileHandle;
 
-static gboolean         ensure_connection       (void);
-static void             append_args_valist      (DBusMessage     *message,
-						 DvdArgumentType  first_arg_type,
-						 va_list          var_args);
-static DBusMessage *    execute_operation       (const gchar     *method,
-						 GnomeVFSContext *context,
-						 GnomeVFSResult  *result,
-						 gint             timeout,
-						 DvdArgumentType  type,
-						 ...);
-static gboolean         check_if_reply_is_error (DBusMessage     *reply,
-						 GnomeVFSResult  *result);
-static DBusMessage *    create_method_call      (const gchar     *method);
-static gint32           cancellation_id_new     (GnomeVFSContext *context);
-static void             cancellation_id_free    (gint32           cancellation_id,
-						 GnomeVFSContext *context);
-static DBusHandlerResult message_handler        (DBusConnection  *conn,
-						 DBusMessage     *message,
-						 gpointer         user_data);
-static gpointer         main_loop_thread        (gpointer         data);
+static gboolean          ensure_connection       (void);
+static void              append_args_valist      (DBusMessage     *message,
+						  DvdArgumentType  first_arg_type,
+						  va_list          var_args);
+static DBusMessage *     execute_operation       (const gchar     *method,
+						  GnomeVFSContext *context,
+						  GnomeVFSResult  *result,
+						  gint             timeout,
+						  DvdArgumentType  type,
+						  ...);
+static gboolean          check_if_reply_is_error (DBusMessage     *reply,
+						  GnomeVFSResult  *result);
+static DBusMessage *     create_method_call      (const gchar     *method);
+static gint32            cancellation_id_new     (GnomeVFSContext *context);
+static void              cancellation_id_free    (gint32           cancellation_id,
+						  GnomeVFSContext *context);
+static DBusHandlerResult message_handler         (DBusConnection  *conn,
+						  DBusMessage     *message,
+						  gpointer         user_data);
+static gpointer          main_loop_thread        (gpointer         data);
+GnomeVFSMethod *         vfs_module_init         (const char      *method_name,
+						  const char      *args);
+void                     vfs_module_shutdown     (GnomeVFSMethod  *method);
+
 
 static DBusConnection *dbus_main_conn = NULL;
 static DBusConnection *dbus_conn = NULL;
 static dbus_int32_t    dbus_conn_id = 0;
 
-static GStaticMutex    mutex = G_STATIC_MUTEX_INIT; 
+static GStaticMutex    mutex = G_STATIC_MUTEX_INIT;
 
 #define MUTEX_LOCK(x) g_static_mutex_lock (&mutex);
 #define MUTEX_UNLOCK(x) g_static_mutex_unlock (&mutex);
@@ -95,13 +102,13 @@ idle_cancel_func (gpointer data)
 {
 	DBusMessage *message;
 	gint32       cancellation_id;
-	
+
 	d(g_print ("Send cancel\n"));
 
 	if (!dbus_connection_get_is_connected (dbus_main_conn)) {
 		return FALSE;
 	}
-	
+
 	message = dbus_message_new_method_call (DVD_DAEMON_SERVICE,
  						DVD_DAEMON_OBJECT,
 						DVD_DAEMON_INTERFACE,
@@ -109,46 +116,40 @@ idle_cancel_func (gpointer data)
 	if (!message) {
 		g_error ("Out of memory");
 	}
-	
+
 	cancellation_id = GPOINTER_TO_INT (data);
-	
+
 	if (!dbus_message_append_args (message,
-				       DBUS_TYPE_INT32, cancellation_id,
-				       DBUS_TYPE_INT32, dbus_conn_id,
+				       DBUS_TYPE_INT32, &cancellation_id,
+				       DBUS_TYPE_INT32, &dbus_conn_id,
 				       DBUS_TYPE_INVALID)) {
 		g_error ("Out of memory");
 	}
-	
+
 	dbus_connection_send (dbus_main_conn, message, NULL);
 	dbus_message_unref (message);
 
 	return FALSE;
 }
 
-static gboolean 
+static gboolean
 ensure_main_connection (void)
 {
-	const gchar *bus_address;
+	DBusError error;
 	
+	/*g_print ("** ensure main connection, thread %p\n", (gpointer) pthread_self ());*/
+				
 	if (dbus_main_conn) {
 		return TRUE;
 	}
 
-	bus_address = g_getenv ("DBUS_SESSION_BUS_ADDRESS");
-	if (!bus_address) {
-		return FALSE;
-	}
-
-	dbus_main_conn = dbus_connection_open (bus_address, NULL);
+	dbus_error_init (&error);
+	dbus_main_conn = dbus_bus_get_private (DBUS_BUS_SESSION, &error);
 	if (!dbus_main_conn) {
-		return FALSE;
-	}
-
-	dbus_connection_set_exit_on_disconnect (dbus_main_conn, TRUE);
-
-	if (!dbus_bus_register (dbus_main_conn, NULL)) {
-		dbus_connection_disconnect (dbus_main_conn);
-		dbus_connection_unref (dbus_main_conn);
+		g_printerr ("Couldn't get main dbus connection: %s\n",
+			    error.message);
+		dbus_error_free (&error);
+		
 		return FALSE;
 	}
 	
@@ -157,7 +158,7 @@ ensure_main_connection (void)
 	return TRUE;
 }
 
-static gboolean 
+static gboolean
 ensure_connection (void)
 {
 	DBusMessage  *message;
@@ -165,16 +166,17 @@ ensure_connection (void)
 	gchar        *address;
 	DBusError     error;
 	GMainContext *context;
-	
+
 	MUTEX_LOCK ("ensure");
 
+	/*g_print ("** ensure connection, thread %p\n", (gpointer) pthread_self ());*/
+	
 	if (dbus_conn && dbus_connection_get_is_connected (dbus_conn)) {
 		MUTEX_UNLOCK ("dbus_conn");
 		return TRUE;
 	}
 
 	if (!ensure_main_connection ()) {
-		g_warning ("Couldn't get main dbus connection.");
 		MUTEX_UNLOCK ("dbus_conn");
 		return FALSE;
 	}
@@ -183,11 +185,9 @@ ensure_connection (void)
 						DVD_DAEMON_OBJECT,
 						DVD_DAEMON_INTERFACE,
 						DVD_DAEMON_METHOD_GET_CONNECTION);
-	dbus_message_set_auto_activation (message, TRUE);
 
 	dbus_error_init (&error);
-
-	reply = dbus_connection_send_with_reply_and_block (dbus_main_conn, 
+	reply = dbus_connection_send_with_reply_and_block (dbus_main_conn,
 							   message,
 							   -1,
 							   &error);
@@ -207,8 +207,8 @@ ensure_connection (void)
 	if (dbus_conn) {
 		dbus_connection_unref (dbus_conn);
 	}
-	
-	dbus_conn = dbus_connection_open (address, &error);
+
+	dbus_conn = dbus_connection_open_private (address, &error);
 	if (!dbus_conn) {
 		g_warning ("Failed to connect to peer-to-peer address (%s): %s",
 			   address, error.message);
@@ -216,8 +216,6 @@ ensure_connection (void)
 		MUTEX_UNLOCK ("p2p conn");
 		return FALSE;
 	}
-      
-	dbus_free (address);
 
 	if (!dbus_connection_add_filter (dbus_conn, message_handler,
 					 NULL, NULL)) {
@@ -231,7 +229,7 @@ ensure_connection (void)
 	}
 
 	context = g_main_context_new ();
-	
+
 	dbus_connection_setup_with_g_main (dbus_conn, context);
 	g_thread_create (main_loop_thread, context, FALSE, NULL);
 
@@ -244,11 +242,11 @@ append_args_valist (DBusMessage     *message,
 		    DvdArgumentType  first_arg_type,
 		    va_list          var_args)
 {
-	DvdArgumentType      type;
-	DBusMessageIter      iter;
+	DvdArgumentType type;
+	DBusMessageIter iter;
+
+	dbus_message_iter_init_append (message, &iter);
 
-	dbus_message_append_iter_init (message, &iter);
-	
 	type = first_arg_type;
 	while (type != DVD_TYPE_LAST) {
 		switch (type) {
@@ -258,11 +256,13 @@ append_args_valist (DBusMessage     *message,
 			gchar       *str;
 
 			uri = va_arg (var_args, GnomeVFSURI *);
-			uri_str = gnome_vfs_uri_to_string (uri, 
+			uri_str = gnome_vfs_uri_to_string (uri,
 							   GNOME_VFS_URI_HIDE_NONE);
 			str = gnome_vfs_escape_host_and_path_string (uri_str);
 			g_free (uri_str);
-			if (!dbus_message_iter_append_string (&iter, str)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_STRING,
+							     &str)) {
 				g_error ("Out of memory");
 			}
 
@@ -273,7 +273,9 @@ append_args_valist (DBusMessage     *message,
 			const gchar *str;
 
 			str = va_arg (var_args, const gchar *);
-			if (!dbus_message_iter_append_string (&iter, str)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_STRING,
+							     &str)) {
 				g_error ("Out of memory");
 			}
 			break;
@@ -282,16 +284,20 @@ append_args_valist (DBusMessage     *message,
 			dbus_int32_t int32;
 
 			int32 = va_arg (var_args, dbus_int32_t);
-			if (!dbus_message_iter_append_int32 (&iter, int32)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_INT32,
+							     &int32)) {
 				g_error ("Out of memory");
 			}
 			break;
 		}
 		case DVD_TYPE_INT64: {
 			dbus_int64_t int64;
-			
+
 			int64 = va_arg (var_args, dbus_int64_t);
-			if (!dbus_message_iter_append_int64 (&iter, int64)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_INT64,
+							     &int64)) {
 				g_error ("Out of memory");
 			}
 			break;
@@ -300,7 +306,9 @@ append_args_valist (DBusMessage     *message,
 			dbus_uint64_t uint64;
 
 			uint64 = va_arg (var_args, dbus_uint64_t);
-			if (!dbus_message_iter_append_uint64 (&iter, uint64)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_UINT64,
+							     &uint64)) {
 				g_error ("Out of memory");
 			}
 			break;
@@ -309,14 +317,16 @@ append_args_valist (DBusMessage     *message,
 			dbus_bool_t bool_v;
 
 			bool_v = va_arg (var_args, dbus_bool_t);
-			if (!dbus_message_iter_append_boolean (&iter, bool_v)) {
+			if (!dbus_message_iter_append_basic (&iter,
+							     DBUS_TYPE_BOOLEAN,
+							     &bool_v)) {
 				g_error ("Out of memory");
 			}
 			break;
 		}
 		case DVD_TYPE_FILE_INFO: {
 			GnomeVFSFileInfo *info;
-			
+
 			info = va_arg (var_args, GnomeVFSFileInfo *);
 			if (!dbus_utils_message_iter_append_file_info (&iter, info)) {
 				g_error ("Out of memory");
@@ -324,13 +334,28 @@ append_args_valist (DBusMessage     *message,
 			break;
 		}
 		case DVD_TYPE_BYTE_ARRAY: {
+			DBusMessageIter      array_iter;
 			unsigned const char *data;
 			gint                 len;
-			
+
 			len = va_arg (var_args, dbus_int32_t);
 			data = va_arg (var_args, unsigned const char *);
+			
+			if (!dbus_message_iter_open_container (&iter,
+							       DBUS_TYPE_ARRAY,
+							       DBUS_TYPE_BYTE_AS_STRING,
+							       &array_iter)) {
+				g_error ("Out of memory");
+			}
 
-			if (!dbus_message_iter_append_byte_array (&iter, data, len)) {
+			if (!dbus_message_iter_append_fixed_array (&array_iter,
+								   DBUS_TYPE_BYTE,
+								   &data,
+								   len)) {
+				g_error ("Out of memory");
+			}
+
+			if (!dbus_message_iter_close_container (&iter, &array_iter)) {
 				g_error ("Out of memory");
 			}
 			break;
@@ -338,7 +363,7 @@ append_args_valist (DBusMessage     *message,
 		case DVD_TYPE_LAST:
 			return;
 		}
-		
+
 		type = va_arg (var_args, DvdArgumentType);
 	}
 }
@@ -348,7 +373,7 @@ execute_operation (const gchar      *method,
 		   GnomeVFSContext  *context,
 		   GnomeVFSResult   *result,
 		   gint              timeout,
-		   DvdArgumentType   first_arg_type, 
+		   DvdArgumentType   first_arg_type,
 		   ...)
 {
 	DBusMessage *message;
@@ -361,12 +386,12 @@ execute_operation (const gchar      *method,
 		if (result) {
 			*result = GNOME_VFS_ERROR_INTERNAL;
 		}
-		
+
 		return NULL;
 	}
-	
+
 	message = create_method_call (method);
-	
+
 	va_start (var_args, first_arg_type);
 	append_args_valist (message, first_arg_type, var_args);
 	va_end (var_args);
@@ -374,11 +399,11 @@ execute_operation (const gchar      *method,
 	cancellation_id = -1;
 	if (context) {
 		cancellation_id = cancellation_id_new (context);
-		dbus_message_append_args (message, 
-					  DBUS_TYPE_INT32, (dbus_int32_t) cancellation_id,
+		dbus_message_append_args (message,
+					  DBUS_TYPE_INT32, &cancellation_id,
 					  DBUS_TYPE_INVALID);
 	}
-	
+
 	dbus_error_init (&error);
 	d(g_print ("Executing operation '%s'... \n", method));
 
@@ -417,7 +442,7 @@ execute_operation (const gchar      *method,
 		if (result) {
 			*result = GNOME_VFS_ERROR_GENERIC;
 		}
-		
+
 		return NULL;
 	}
 
@@ -436,10 +461,10 @@ check_if_reply_is_error (DBusMessage *reply, GnomeVFSResult *result)
 	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &r,
 			       DBUS_TYPE_INVALID);
-	
+
 	d(g_print ("check_if_reply_is_error: %s\n",
 		   gnome_vfs_result_to_string (r)));
-	
+
 	if (result) {
 		*result = r;
 	}
@@ -457,7 +482,7 @@ static DBusMessage *
 create_method_call (const gchar *method)
 {
 	DBusMessage *message;
-	
+
 	message = dbus_message_new_method_call (DVD_DAEMON_SERVICE,
                                                 DVD_DAEMON_OBJECT,
                                                 DVD_DAEMON_INTERFACE,
@@ -489,7 +514,7 @@ directory_handle_free (DirectoryHandle *handle)
 }
 
 static FileHandle *
-file_handle_new (gint32 handle_id) 
+file_handle_new (gint32 handle_id)
 {
 	FileHandle *handle;
 
@@ -527,7 +552,7 @@ cancellation_id_new (GnomeVFSContext *context)
 				GINT_TO_POINTER (next_handle));
 		}
 	}
-	
+
 	return next_handle;
 }
 
@@ -549,16 +574,8 @@ cancellation_id_free (gint32           cancellation_id,
 	 */
 }
 
-/* This is in gnome-vfs-monitor-private.h which is not installed, but it is
- * still part of the module API, so we need to have this hack here until that is
- * rectified in GnomeVFS.
- */
-void gnome_vfs_monitor_callback (GnomeVFSMethodHandle *method_handle,
-				 GnomeVFSURI *info_uri,
-				 GnomeVFSMonitorEventType event_type);
-
 /* This is a filter function, so never return HANDLED. */
-static DBusHandlerResult 
+static DBusHandlerResult
 message_handler (DBusConnection *conn,
 		 DBusMessage    *message,
 		 gpointer        user_data)
@@ -569,7 +586,7 @@ message_handler (DBusConnection *conn,
 	GnomeVFSURI  *uri;
 
 	/* Check if signal */
-	if (!dbus_message_is_signal (message, DVD_DAEMON_INTERFACE, 
+	if (!dbus_message_is_signal (message, DVD_DAEMON_INTERFACE,
 				     DVD_DAEMON_MONITOR_SIGNAL)) {
 		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 	}
@@ -581,10 +598,9 @@ message_handler (DBusConnection *conn,
 			       DBUS_TYPE_INVALID);
 
 	uri = gnome_vfs_uri_new (uri_str);
-	dbus_free (uri_str);
-	
+
 	gnome_vfs_monitor_callback ((GnomeVFSMethodHandle *) GINT_TO_POINTER (id),
-				    uri, 
+				    uri,
 				    (GnomeVFSMonitorEventType) event_type);
 
 	gnome_vfs_uri_unref (uri);
@@ -597,6 +613,8 @@ main_loop_thread (gpointer data)
 {
 	GMainContext *context = (GMainContext *) data;
 
+	/*g_print ("main_loop_thread, thread %p", (gpointer) pthread_self ());*/
+	
 	while (TRUE) {
 		MUTEX_LOCK ("iterating");
 		g_main_context_iteration (context, FALSE);
@@ -605,9 +623,9 @@ main_loop_thread (gpointer data)
 	}
 
 	return NULL;
-}	
+}
 
-static GnomeVFSResult 
+static GnomeVFSResult
 do_open (GnomeVFSMethod        *method,
 	 GnomeVFSMethodHandle **method_handle,
 	 GnomeVFSURI           *uri,
@@ -634,18 +652,18 @@ do_open (GnomeVFSMethod        *method,
 		return result;
 	}
 
-	dbus_message_get_args (reply, NULL, 
+	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_INT32, &handle_id,
 			       DBUS_TYPE_INVALID);
-	
+
 	handle = file_handle_new (handle_id);
 
 	*method_handle = (GnomeVFSMethodHandle *) handle;
 
 	dbus_message_unref (reply);
-	
-	return result; 
+
+	return result;
 }
 
 static GnomeVFSResult
@@ -682,7 +700,7 @@ do_create (GnomeVFSMethod        *method,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_INT32, &handle_id,
 			       DBUS_TYPE_INVALID);
-	
+
 	handle = file_handle_new (handle_id);
 
 	*method_handle = (GnomeVFSMethodHandle *) handle;
@@ -704,7 +722,7 @@ do_close (GnomeVFSMethod       *method,
 	handle = (FileHandle *) method_handle;
 
 	reply = execute_operation (DVD_DAEMON_METHOD_CLOSE,
-				   context, &result, 
+				   context, &result,
 				   DBUS_TIMEOUT_OPEN_CLOSE,
 				   DVD_TYPE_INT32, handle->handle_id,
 				   DVD_TYPE_LAST);
@@ -716,11 +734,11 @@ do_close (GnomeVFSMethod       *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
 
 	file_handle_free (handle);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -732,14 +750,15 @@ do_read (GnomeVFSMethod *method,
 	 GnomeVFSFileSize *bytes_read,
 	 GnomeVFSContext *context)
 {
-	FileHandle       *handle;
-	DBusMessage      *reply;
-	GnomeVFSResult    result;
-	gint              size;
-	unsigned char    *data;
-	DBusMessageIter   iter;
-	int               type;
-	
+	FileHandle      *handle;
+	DBusMessage     *reply;
+	GnomeVFSResult   result;
+	gint             size;
+	guchar          *data;
+	DBusMessageIter  iter;
+	DBusMessageIter  array_iter;
+	int              type;
+
 	handle = (FileHandle *) method_handle;
 
 	reply = execute_operation (DVD_DAEMON_METHOD_READ,
@@ -760,24 +779,25 @@ do_read (GnomeVFSMethod *method,
 
 	dbus_message_iter_init (reply, &iter);
 
-	dbus_message_iter_next (&iter); /* Not interested in result */
-	
+	dbus_message_iter_next (&iter); /* Result is already checked. */
+
 	type = dbus_message_iter_get_arg_type (&iter);
 	if (type != DBUS_TYPE_ARRAY) {
-		/* FIXME: Error */
+		dbus_message_unref (reply);
+		return GNOME_VFS_ERROR_INTERNAL;
 	}
 
-	if (!dbus_message_iter_get_byte_array (&iter, &data, &size)) {
-		/* FIXME: Error */
-	}
+	dbus_message_iter_recurse (&iter, &array_iter);
+	if (dbus_message_iter_get_array_len (&array_iter) > 0) {
+		dbus_message_iter_get_fixed_array (&array_iter, &data, &size);
+		
+		memcpy (buffer, data, size);
+	} 
+
+	*bytes_read = size;
 
 	dbus_message_unref (reply);
-	
-	memcpy (buffer, data, size);
-	*bytes_read = (GnomeVFSFileSize) size;
 
-	dbus_free (data);
-	
 	return GNOME_VFS_OK;
 }
 
@@ -793,13 +813,13 @@ do_write (GnomeVFSMethod *method,
 	DBusMessage      *reply;
 	GnomeVFSResult    result;
 	dbus_uint64_t     written;
-	
+
 	handle = (FileHandle *) method_handle;
 
 	reply = execute_operation (DVD_DAEMON_METHOD_WRITE,
 				   context, &result, -1,
 				   DVD_TYPE_INT32, handle->handle_id,
-				   DVD_TYPE_BYTE_ARRAY, (dbus_int32_t) num_bytes, 
+				   DVD_TYPE_BYTE_ARRAY, (dbus_int32_t) num_bytes,
 				   (const unsigned char *) buffer,
 				   DVD_TYPE_LAST);
 
@@ -810,7 +830,7 @@ do_write (GnomeVFSMethod *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_UINT64, &written,
@@ -820,7 +840,7 @@ do_write (GnomeVFSMethod *method,
 
 	*bytes_written = written;
 
-	return GNOME_VFS_OK; 
+	return GNOME_VFS_OK;
 }
 
 static GnomeVFSResult
@@ -858,7 +878,7 @@ do_seek (GnomeVFSMethod *method,
 static GnomeVFSResult
 do_tell (GnomeVFSMethod *method,
 	 GnomeVFSMethodHandle *method_handle,
-	 GnomeVFSFileOffset *offset_return)
+	 GnomeVFSFileSize *offset_return)
 {
 	FileHandle     *handle;
         DBusMessage    *reply;
@@ -879,13 +899,13 @@ do_tell (GnomeVFSMethod *method,
 		return result;
 	}
 
-	dbus_message_get_args (reply, NULL, 
+	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_INT64, &offset,
 			       DBUS_TYPE_INVALID);
 
 	*offset_return = offset;
-	
+
 	dbus_message_unref (reply);
 
 	return GNOME_VFS_OK;
@@ -916,7 +936,7 @@ do_truncate_handle (GnomeVFSMethod *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
 
 	return GNOME_VFS_OK;
@@ -945,22 +965,22 @@ do_open_directory (GnomeVFSMethod           *method,
 	if (!reply) {
 		return result;
 	}
-	
+
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_INT32, &handle_id,
 			       DBUS_TYPE_INVALID);
-	
+
 	handle = directory_handle_new (handle_id);
 
 	*method_handle = (GnomeVFSMethodHandle *) handle;
-	
+
 	dbus_message_unref (reply);
-	
+
 	return result;
 }
 
@@ -987,11 +1007,11 @@ do_close_directory (GnomeVFSMethod       *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
 
 	directory_handle_free (handle);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1021,9 +1041,11 @@ do_read_directory (GnomeVFSMethod       *method,
 		}
 
 		if (check_if_reply_is_error (reply, &result)) {
-			return result;
+			if (result != GNOME_VFS_ERROR_EOF) {
+				return result;
+			}
 		}
-		
+
 		list = dbus_utils_message_get_file_info_list (reply);
 
 		dbus_message_unref (reply);
@@ -1032,21 +1054,31 @@ do_read_directory (GnomeVFSMethod       *method,
 		handle->current = list;
 
 		d(g_print ("got list: %d\n", g_list_length (list)));
-		
+
+		/* If we get OK, and an empty list, it means that there are no
+		 * more files.
+		 */
 		if (list == NULL) {
 			return GNOME_VFS_ERROR_EOF;
 		}
 	}
-	
+
 	if (handle->current) {
 		file_info_src = handle->current->data;
 		gnome_vfs_file_info_copy (file_info, file_info_src);
 
 		handle->current = handle->current->next;
+
+		/* If this was the last in this chunk, read more files in the
+		 * next call.
+		 */
+		if (!handle->current) {
+			handle->dirs = NULL;
+		}
 	} else {
 		return GNOME_VFS_ERROR_EOF;
 	}
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1055,9 +1087,9 @@ get_file_info_from_message (DBusMessage *message)
 {
 	GnomeVFSFileInfo *info;
 	DBusMessageIter   iter;
-		
+
 	dbus_message_iter_init (message, &iter);
-	
+
 	/* Not interested in result */
 	dbus_message_iter_next (&iter);
 
@@ -1079,7 +1111,7 @@ do_get_file_info (GnomeVFSMethod *method,
         DBusMessage       *reply;
 	GnomeVFSResult     result;
 	GnomeVFSFileInfo  *info;
-	
+
 	reply = execute_operation (DVD_DAEMON_METHOD_GET_FILE_INFO,
 				   context, &result, -1,
 				   DVD_TYPE_URI, uri,
@@ -1093,7 +1125,7 @@ do_get_file_info (GnomeVFSMethod *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	info = get_file_info_from_message (reply);
 	dbus_message_unref (reply);
 
@@ -1118,9 +1150,9 @@ do_get_file_info_from_handle (GnomeVFSMethod *method,
         DBusMessage       *reply;
 	GnomeVFSResult     result;
 	GnomeVFSFileInfo  *info;
-	
+
 	handle = (FileHandle *) method_handle;
-	
+
 	reply = execute_operation (DVD_DAEMON_METHOD_GET_FILE_INFO_FROM_HANDLE,
 				   context, &result, -1,
 				   DVD_TYPE_INT32, handle->handle_id,
@@ -1149,7 +1181,7 @@ do_get_file_info_from_handle (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-static gboolean 
+static gboolean
 do_is_local (GnomeVFSMethod *method, const GnomeVFSURI *uri)
 {
         DBusMessage    *reply;
@@ -1160,22 +1192,22 @@ do_is_local (GnomeVFSMethod *method, const GnomeVFSURI *uri)
 				   NULL, &result, -1,
 				   DVD_TYPE_URI, uri,
 				   DVD_TYPE_LAST);
-	
+
 	if (!reply) {
 		return result;
 	}
-	
+
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_BOOLEAN, &is_local,
 			       DBUS_TYPE_INVALID);
-	
+
 	dbus_message_unref (reply);
-	
+
 	return is_local;
 }
 
@@ -1203,9 +1235,9 @@ do_make_directory (GnomeVFSMethod  *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1218,7 +1250,7 @@ do_remove_directory (GnomeVFSMethod  *method,
 	GnomeVFSResult  result;
 
 	/*g_print ("thread: %p", (gpointer) pthread_self ());*/
-	
+
 	reply = execute_operation (DVD_DAEMON_METHOD_REMOVE_DIRECTORY,
 				   context, &result, -1,
 				   DVD_TYPE_URI, uri,
@@ -1231,9 +1263,9 @@ do_remove_directory (GnomeVFSMethod  *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1255,17 +1287,17 @@ do_move (GnomeVFSMethod  *method,
 				   DVD_TYPE_URI, new_uri,
 				   DVD_TYPE_BOOL, force_replace,
 				   DVD_TYPE_LAST);
-	
+
 	if (!reply) {
 		return result;
 	}
-	
+
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1287,13 +1319,13 @@ do_unlink (GnomeVFSMethod  *method,
 	if (!reply) {
 		return result;
 	}
-	
+
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1307,7 +1339,7 @@ do_check_same_fs (GnomeVFSMethod  *method,
 	DBusMessage    *reply;
 	GnomeVFSResult  result;
 	dbus_bool_t     same_fs;
-	
+
 	reply = execute_operation (DVD_DAEMON_METHOD_CHECK_SAME_FS,
 				   context, &result, -1,
 				   DVD_TYPE_URI, a,
@@ -1321,7 +1353,7 @@ do_check_same_fs (GnomeVFSMethod  *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_get_args (reply, NULL,
 			       DBUS_TYPE_INT32, &result,
 			       DBUS_TYPE_BOOLEAN, &same_fs,
@@ -1365,10 +1397,10 @@ do_set_file_info (GnomeVFSMethod          *method,
 }
 
 static GnomeVFSResult
-do_truncate (GnomeVFSMethod *method,
-	     GnomeVFSURI *uri,
-	     GnomeVFSFileSize where,
-	     GnomeVFSContext *context)
+do_truncate (GnomeVFSMethod   *method,
+	     GnomeVFSURI       *uri,
+	     GnomeVFSFileSize  where,
+	     GnomeVFSContext  *context)
 {
 	DBusMessage    *reply;
 	GnomeVFSResult  result;
@@ -1393,14 +1425,14 @@ do_truncate (GnomeVFSMethod *method,
 }
 
 static GnomeVFSResult
-do_find_directory (GnomeVFSMethod *method,
-		   GnomeVFSURI *near_uri,
-		   GnomeVFSFindDirectoryKind kind,
-		   GnomeVFSURI **result_uri,
-		   gboolean create_if_needed,
-		   gboolean find_if_needed,
-		   guint permissions,
-		   GnomeVFSContext *context)
+do_find_directory (GnomeVFSMethod             *method,
+		   GnomeVFSURI                *near_uri,
+		   GnomeVFSFindDirectoryKind   kind,
+		   GnomeVFSURI               **result_uri,
+		   gboolean                    create_if_needed,
+		   gboolean                    find_if_needed,
+		   guint                       permissions,
+		   GnomeVFSContext            *context)
 {
 	DBusMessage    *reply;
 	GnomeVFSResult  result;
@@ -1430,9 +1462,7 @@ do_find_directory (GnomeVFSMethod *method,
 	dbus_message_unref (reply);
 
 	*result_uri = gnome_vfs_uri_new (uri_str);
-	
-	dbus_free (uri_str);
-	
+
 	return GNOME_VFS_OK;
 }
 
@@ -1458,7 +1488,7 @@ do_create_symbolic_link (GnomeVFSMethod  *method,
 	if (check_if_reply_is_error (reply, &result)) {
 		return result;
 	}
-	
+
 	dbus_message_unref (reply);
 
 	return GNOME_VFS_OK;
@@ -1473,7 +1503,7 @@ do_monitor_add (GnomeVFSMethod        *method,
 	DBusMessage    *reply;
 	GnomeVFSResult  result;
 	dbus_int32_t    id;
-	
+
 	reply = execute_operation (DVD_DAEMON_METHOD_MONITOR_ADD,
 				   NULL, &result, -1,
 				   DVD_TYPE_URI, uri,
@@ -1494,7 +1524,7 @@ do_monitor_add (GnomeVFSMethod        *method,
 			       DBUS_TYPE_INVALID);
 
 	*method_handle = GINT_TO_POINTER (id);
-	
+
 	dbus_message_unref (reply);
 
 	return GNOME_VFS_OK;
@@ -1525,9 +1555,75 @@ do_monitor_cancel (GnomeVFSMethod       *method,
 	return GNOME_VFS_OK;
 }
 
+static GnomeVFSResult
+do_forget_cache	(GnomeVFSMethod       *method,
+		 GnomeVFSMethodHandle *method_handle,
+		 GnomeVFSFileOffset    offset,
+		 GnomeVFSFileSize      size)
+{
+	FileHandle     *handle;
+	DBusMessage    *reply;
+	GnomeVFSResult  result;
+
+	handle = (FileHandle *) method_handle;
+
+	reply = execute_operation (DVD_DAEMON_METHOD_FORGET_CACHE,
+				   NULL, &result, -1,
+				   DVD_TYPE_INT32, handle->handle_id,
+				   DVD_TYPE_INT64, offset,
+				   DVD_TYPE_UINT64, size,
+				   DVD_TYPE_LAST);
+
+	if (!reply) {
+		return result;
+	}
+
+	if (check_if_reply_is_error (reply, &result)) {
+		return result;
+	}
+
+	dbus_message_unref (reply);
+
+	return GNOME_VFS_OK;
+}
+
+static GnomeVFSResult
+do_get_volume_free_space (GnomeVFSMethod    *method,
+			  const GnomeVFSURI *uri,
+			  GnomeVFSFileSize  *free_space)
+{
+	DBusMessage    *reply;
+	GnomeVFSResult  result;
+	dbus_uint64_t    space;
+
+	reply = execute_operation (DVD_DAEMON_METHOD_GET_VOLUME_FREE_SPACE,
+				   NULL, &result, -1,
+				   DVD_TYPE_URI, uri,
+				   DVD_TYPE_LAST);
+
+	if (!reply) {
+		return result;
+	}
+
+	if (check_if_reply_is_error (reply, &result)) {
+		return result;
+	}
+
+	dbus_message_get_args (reply, NULL,
+			       DBUS_TYPE_INT32, &result,
+			       DBUS_TYPE_UINT64, &space,
+			       DBUS_TYPE_INVALID);
+
+	*free_space = space;
+
+	dbus_message_unref (reply);
+
+	return GNOME_VFS_OK;
+}
+
 static GnomeVFSMethod method = {
         sizeof (GnomeVFSMethod),
-                                                                                
+
         do_open,                      /* open */
         do_create,                    /* create */
         do_close,                     /* close */
@@ -1553,16 +1649,16 @@ static GnomeVFSMethod method = {
         do_create_symbolic_link,      /* create_symbolic_link */
 	do_monitor_add,               /* monitor_add */
         do_monitor_cancel,            /* monitor_cancel */
-	NULL                          /* file_control */
+	NULL,                         /* file_control */
+	do_forget_cache,              /* forget_cache */
+	do_get_volume_free_space      /* get_volume_free_space */
 };
 
-GnomeVFSMethod *vfs_module_init     (const char     *method_name,
-				     const char     *args);
-void            vfs_module_shutdown (GnomeVFSMethod *method);
-
 GnomeVFSMethod *
 vfs_module_init (const char *method_name, const char *args)
 {
+	/*g_print ("** initing dbus method, thread %p\n", (gpointer) pthread_self ());*/
+
 	return &method;
 }
 
@@ -1572,7 +1668,7 @@ vfs_module_shutdown (GnomeVFSMethod* method)
 	if (dbus_conn) {
 		dbus_connection_disconnect (dbus_conn);
 		dbus_connection_unref (dbus_conn);
-		
+
 		dbus_conn = NULL;
 	}
 }
