@@ -849,7 +849,7 @@ gnome_vfs_module_callback_invoke (const char    *callback_name,
 	callback = NULL;
 
 	if (gnome_vfs_get_is_daemon()) {
-#ifndef USE_EXT_DBUS_DAEMON
+#ifndef USE_DBUS_DAEMON
 		return _gnome_vfs_module_callback_marshal_invoke (callback_name,
 								  in, in_size,
 								  out, out_size);
