@@ -26,7 +26,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
-#include <libgnomevfs/gnome-vfs-i18n.h>
+#include <glib/gi18n-lib.h>
 
 #include "gnome-vfs-volume-monitor-daemon.h"
 #include "gnome-vfs-volume-monitor-private.h"
@@ -63,7 +63,9 @@ static GnomeVFSVolumeMonitorClass *parent_class = NULL;
  * This is also useful for maintaing the non-HAL code on a system with
  * HAL installed.
  */
+#ifdef USE_HAL
 static gboolean dont_use_hald = TRUE;
+#endif /* USE_HAL */
 
 /* Locking strategy:
  *
@@ -153,8 +155,10 @@ connected_servers_changed (GConfClient* client,
 static void
 gnome_vfs_volume_monitor_daemon_init (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 {
+	volume_monitor_daemon->gconf_client = gconf_client_get_default ();
+
 #ifdef USE_HAL
-	if (_gnome_vfs_monitor_hal_mounts_init (volume_monitor_daemon)) {
+	if (_gnome_vfs_hal_mounts_init (volume_monitor_daemon)) {
 		/* It worked, do use HAL */
 		dont_use_hald = FALSE;
 	} else {
@@ -163,14 +167,11 @@ gnome_vfs_volume_monitor_daemon_init (GnomeVFSVolumeMonitorDaemon *volume_monito
 	}
 #endif /* USE_HAL */
 
-	if (dont_use_hald) {
-		_gnome_vfs_monitor_unix_mounts (fstab_changed,
-						volume_monitor_daemon,
-						mtab_changed,
-						volume_monitor_daemon);
-	}
+	_gnome_vfs_monitor_unix_mounts (fstab_changed,
+					volume_monitor_daemon,
+					mtab_changed,
+					volume_monitor_daemon);
 
-	volume_monitor_daemon->gconf_client = gconf_client_get_default ();
 	gconf_client_add_dir (volume_monitor_daemon->gconf_client,
 			      CONNECTED_SERVERS_DIR,
 			      GCONF_CLIENT_PRELOAD_RECURSIVE,
@@ -185,16 +186,8 @@ gnome_vfs_volume_monitor_daemon_init (GnomeVFSVolumeMonitorDaemon *volume_monito
 					 NULL);
 								       
 	
-	if (dont_use_hald) {
-		update_fstab_drives (volume_monitor_daemon);
-		update_mtab_volumes (volume_monitor_daemon);
-	}
-#ifdef USE_HAL
-	else {
-		_gnome_vfs_monitor_hal_get_volume_list (volume_monitor_daemon);
-	}
-#endif /* USE_HAL */
-
+	update_fstab_drives (volume_monitor_daemon);
+	update_mtab_volumes (volume_monitor_daemon);
 	update_connected_servers (volume_monitor_daemon);
 }
 
@@ -205,15 +198,12 @@ gnome_vfs_volume_monitor_daemon_force_probe (GnomeVFSVolumeMonitor *volume_monit
 	
 	volume_monitor_daemon = GNOME_VFS_VOLUME_MONITOR_DAEMON (volume_monitor);
 	
-	if (dont_use_hald) {
-		update_fstab_drives (volume_monitor_daemon);
-		update_mtab_volumes (volume_monitor_daemon);
-	}
 #ifdef USE_HAL
-	else {
-		_gnome_vfs_monitor_hal_get_volume_list (volume_monitor_daemon);
-	}
-#endif /* USE_HAL */
+	if (!dont_use_hald)
+		_gnome_vfs_hal_mounts_force_reprobe (volume_monitor_daemon);
+#endif
+	update_fstab_drives (volume_monitor_daemon);
+	update_mtab_volumes (volume_monitor_daemon);
 	update_connected_servers (volume_monitor_daemon);
 }
 
@@ -225,27 +215,26 @@ gnome_vfs_volume_monitor_daemon_finalize (GObject *object)
 	GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon;
 
 	volume_monitor_daemon = GNOME_VFS_VOLUME_MONITOR_DAEMON (object);
-		
-	if (dont_use_hald) {
-		_gnome_vfs_stop_monitoring_unix_mounts ();
-
-		g_list_foreach (volume_monitor_daemon->last_mtab,
-				(GFunc)_gnome_vfs_unix_mount_free, NULL);
-		g_list_free (volume_monitor_daemon->last_mtab);
 	
-		g_list_foreach (volume_monitor_daemon->last_fstab,
-				(GFunc)_gnome_vfs_unix_mount_point_free, NULL);
-		g_list_free (volume_monitor_daemon->last_fstab);	
-	}
-#ifdef USE_HAL
-	else {
-		_gnome_vfs_monitor_hal_mounts_shutdown (volume_monitor_daemon);
-	}
-#endif /* USE_HAL */
+	_gnome_vfs_stop_monitoring_unix_mounts ();
+	
+	g_list_foreach (volume_monitor_daemon->last_mtab,
+			(GFunc)_gnome_vfs_unix_mount_free, NULL);
+	g_list_free (volume_monitor_daemon->last_mtab);
+
+	g_list_foreach (volume_monitor_daemon->last_fstab,
+			(GFunc)_gnome_vfs_unix_mount_point_free, NULL);
+	g_list_free (volume_monitor_daemon->last_fstab);
 
 	gconf_client_notify_remove (volume_monitor_daemon->gconf_client,
 				    volume_monitor_daemon->connected_id);
 
+#ifdef USE_HAL
+	if (!dont_use_hald) {
+		_gnome_vfs_hal_mounts_shutdown (volume_monitor_daemon);
+	}
+#endif /* USE_HAL */
+
 	g_object_unref (volume_monitor_daemon->gconf_client);
 	
 	if (G_OBJECT_CLASS (parent_class)->finalize)
@@ -743,7 +732,13 @@ update_fstab_drives (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 			drive = _gnome_vfs_volume_monitor_find_fstab_drive_by_activation_uri (volume_monitor, uri);
 
 			if (drive != NULL) {
+#ifdef USE_HAL
+				/* don't remove if managed by the hal backend */
+				if (dont_use_hald || drive->priv->hal_udi == NULL)
+					_gnome_vfs_volume_monitor_disconnected (volume_monitor, drive);
+#else
 				_gnome_vfs_volume_monitor_disconnected (volume_monitor, drive);
+#endif
 			} else {
 				g_warning ("removed drive not in old fstab list??");
 			}
@@ -755,12 +750,30 @@ update_fstab_drives (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 			mount = l->data;
 		
 			drive = create_drive_from_mount_point (volume_monitor, mount);
+
 			if (drive != NULL) {
+
+#ifdef USE_HAL
+				if (!dont_use_hald) {
+					GnomeVFSDrive *new_drive;
+
+					new_drive = _gnome_vfs_hal_mounts_modify_drive (volume_monitor_daemon, drive);
+					if (new_drive != NULL) {
+						_gnome_vfs_volume_monitor_connected (volume_monitor, new_drive);
+						gnome_vfs_drive_unref (new_drive);
+					}
+				} else {
+					_gnome_vfs_volume_monitor_connected (volume_monitor, drive);
+					gnome_vfs_drive_unref (drive);
+				}
+#else
 				_gnome_vfs_volume_monitor_connected (volume_monitor, drive);
 				gnome_vfs_drive_unref (drive);
+#endif /* USE_HAL */
+					
 			}
 		}
-		
+
 		g_list_free (added);
 		g_list_free (removed);
 		g_list_foreach (volume_monitor_daemon->last_fstab,
@@ -1008,7 +1021,13 @@ update_mtab_volumes (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 			vol = _gnome_vfs_volume_monitor_find_mtab_volume_by_activation_uri (volume_monitor, uri);
 
 			if (vol != NULL) {
+#ifdef USE_HAL
+				/* don't remove if managed by the hal backend */
+				if (dont_use_hald || vol->priv->hal_udi == NULL)
+					_gnome_vfs_volume_monitor_unmounted (volume_monitor, vol);
+#else
 				_gnome_vfs_volume_monitor_unmounted (volume_monitor, vol);
+#endif
 			} else {
 				g_warning ("removed volume not in old mtab list??");
 			}
@@ -1026,8 +1045,23 @@ update_mtab_volumes (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 		
 			vol = create_vol_from_mount (volume_monitor, mount);
 			vol->priv->unix_device = unix_device;
-			_gnome_vfs_volume_monitor_mounted (volume_monitor, vol);
+#ifdef USE_HAL
+			if (!dont_use_hald) {
+				GnomeVFSVolume *new_vol;
+
+				new_vol = _gnome_vfs_hal_mounts_modify_volume (volume_monitor_daemon, vol);
+				if (new_vol != NULL) {
+					_gnome_vfs_volume_monitor_mounted (volume_monitor, new_vol);
+					gnome_vfs_volume_unref (new_vol);
+				}
+			} else {
+				_gnome_vfs_volume_monitor_mounted (volume_monitor, vol);			
+				gnome_vfs_volume_unref (vol);
+			}
+#else
+			_gnome_vfs_volume_monitor_mounted (volume_monitor, vol);			
 			gnome_vfs_volume_unref (vol);
+#endif /* USE_HAL */
 		}
 
 		g_list_free (devices);
@@ -1038,8 +1072,6 @@ update_mtab_volumes (GnomeVFSVolumeMonitorDaemon *volume_monitor_daemon)
 		g_list_free (volume_monitor_daemon->last_mtab);
 		volume_monitor_daemon->last_mtab = new_mtab;
 	}
-
-	
 }
 
 /************************* connected server ***********************************/
