@@ -25,20 +25,24 @@
 
 #include <string.h>
 #include <stdlib.h>
+#include <glib.h>
+#include <glib/gi18n-lib.h>
+#ifndef G_OS_WIN32
 #include <pthread.h>
+#endif
 
-#ifndef USE_EXT_DBUS_DAEMON
-#include <libbonobo.h>
+#ifndef USE_DBUS_DAEMON
+#include <bonobo/bonobo-exception.h>
+#include <bonobo/bonobo-object.h>
 #endif
 
 #include <gconf/gconf-client.h>
-#include "gnome-vfs-i18n.h"
 #include "gnome-vfs-volume-monitor-private.h"
 #include "gnome-vfs-volume.h"
 #include "gnome-vfs-utils.h"
 #include "gnome-vfs-drive.h"
 
-#ifndef USE_EXT_DBUS_DAEMON
+#ifndef USE_DBUS_DAEMON
 #include "gnome-vfs-client.h"
 #else
 #include "gnome-vfs-volume-monitor-daemon.h"
@@ -47,6 +51,12 @@
 
 #include "gnome-vfs-private.h"
 
+#ifdef USE_HAL
+#include "gnome-vfs-hal-mounts.h"
+#endif
+
+#ifndef G_OS_WIN32
+
 #ifdef USE_VOLRMMOUNT
 
 static const char *volrmmount_locations [] = {
@@ -68,6 +78,7 @@ static const char *mount_known_locations [] = {
 };
 
 static const char *umount_known_locations [] = {
+	"/usr/bin/pumount", "/bin/pumount",
 	"/sbin/umount", "/bin/umount",
 	"/usr/sbin/umount", "/usr/bin/umount",
 	NULL
@@ -87,7 +98,7 @@ find_command (const char **known_locations)
 	int i;
 
 	for (i = 0; known_locations [i]; i++){
-		if (g_file_test (known_locations [i], G_FILE_TEST_EXISTS))
+		if (g_file_test (known_locations [i], G_FILE_TEST_IS_EXECUTABLE))
 			return known_locations [i];
 	}
 	return NULL;
@@ -97,8 +108,10 @@ typedef struct {
 	char *argv[4];
 	char *mount_point;
 	char *device_path;
+	char *hal_udi;
 	GnomeVFSDeviceType device_type;
 	gboolean should_mount;
+	gboolean should_unmount;
 	gboolean should_eject;
 	GnomeVFSVolumeOpCallback callback;
 	gpointer user_data;
@@ -160,7 +173,7 @@ static void
 force_probe (void)
 {
 	GnomeVFSVolumeMonitor *volume_monitor;
-#ifndef USE_EXT_DBUS_DAEMON
+#ifndef USE_DBUS_DAEMON
 	GnomeVFSClient *client;
 	GNOME_VFS_Daemon daemon;
 	CORBA_Environment  ev;
@@ -173,7 +186,7 @@ force_probe (void)
 		callback = _gnome_vfs_get_daemon_force_probe_callback();
 		(*callback) (GNOME_VFS_VOLUME_MONITOR (volume_monitor));
 	} else {
-#ifndef USE_EXT_DBUS_DAEMON
+#ifndef USE_DBUS_DAEMON
 		client = _gnome_vfs_get_client ();
 		daemon = _gnome_vfs_client_get_daemon (client);
 
@@ -218,7 +231,7 @@ report_mount_result (gpointer callback_data)
 		i++;
 	}
 
-#ifdef USE_EXT_DBUS_DAEMON
+#ifdef USE_DBUS_DAEMON
 	/* RH, send off a fake CHANGED event on mtab here so that the FAM-less
 	 * vfs daemon can pick up the change and notify clients about the
 	 * mounted/unmounted volume.
@@ -243,6 +256,7 @@ report_mount_result (gpointer callback_data)
 	
 	g_free (info->mount_point);
 	g_free (info->device_path);
+	g_free (info->hal_udi);
 	g_free (info->error_message);
 	g_free (info->detailed_error_message);
 	g_free (info);
@@ -264,31 +278,51 @@ mount_unmount_thread (void *arg)
 
 	info = arg;
 
-	if (info != NULL) {
+	info->succeeded = TRUE;
+	info->error_message = NULL;
+	info->detailed_error_message = NULL;
+
+	if (info->should_mount || info->should_unmount) {
 		error = NULL;
 		if (g_spawn_sync (NULL,
-				   info->argv,
+				  info->argv,
+#if defined(USE_HAL) && defined(HAL_MOUNT) && defined(HAL_UMOUNT)
+				  /* do pass our environment when using hal mount progams */
+				  ((strcmp (info->argv[0], HAL_MOUNT) == 0) ||
+				   (strcmp (info->argv[0], HAL_UMOUNT) == 0)) ? NULL : envp,
+#else
 				   envp,
+#endif
 				   G_SPAWN_STDOUT_TO_DEV_NULL,
 				   NULL, NULL,
 				   NULL,
 				   &standard_error,
 				   &exit_status,
 				   &error)) {
-			if (exit_status == 0) {
-				info->succeeded = TRUE;
-				info->error_message = NULL;
-				info->detailed_error_message = NULL;
-			} else {
+			if (exit_status != 0) {
 				info->succeeded = FALSE;
-				if (info->should_mount) {
-					info->error_message = generate_mount_error_message (standard_error,
-											    info->device_type);
+				if (strlen (standard_error) > 0) {
+					if (info->should_mount) {
+						info->error_message = generate_mount_error_message (standard_error,
+												    info->device_type);
+					} else {
+						info->error_message = generate_unmount_error_message (standard_error,
+												      info->device_type);
+					}
+					info->detailed_error_message = g_strdup (standard_error);
 				} else {
-					info->error_message = generate_unmount_error_message (standard_error,
-											      info->device_type);
+					/* As of 2.12 we introduce a new contract between gnome-vfs clients
+					 * invoking mount/unmount and the gnome-vfs-daemon:
+					 *
+					 *       "don't display an error dialog if error_message and
+					 *        detailed_error_message are both empty strings".
+					 *
+					 * We want this as we may use mount/unmount/ejects programs that
+					 * shows it's own dialogs. 
+					 */
+					info->error_message = g_strdup ("");
+					info->detailed_error_message = g_strdup ("");
 				}
-				info->detailed_error_message = g_strdup (standard_error);
 			}
 
 			g_free (standard_error);
@@ -302,24 +336,36 @@ mount_unmount_thread (void *arg)
 	}
 
 	if (info->should_eject) {
-#ifdef __FreeBSD__
-	    	char *argv[5] = {
-		    	"cdcontrol",
-			"-f",
-			NULL,
-			"eject",
-			NULL
-		};
-		argv[2] = info->device_path?info->device_path:info->mount_point;
-#else
-		char *argv[3] = {
-			"eject",
-			NULL,
-			NULL
-		};
-		argv[1] = info->device_path?info->device_path:info->mount_point;
+		char *argv[5];
+
+		argv[0] = NULL;
+
+#if defined(USE_HAL) && defined(HAL_EJECT)
+		if (info->hal_udi != NULL) {
+			argv[0] = HAL_EJECT;
+			argv[1] = info->device_path;
+			argv[2] = NULL;
+			
+			if (!g_file_test (argv [0], G_FILE_TEST_IS_EXECUTABLE))
+				argv[0] = NULL;
+		}
 #endif
 
+		if (argv[0] == NULL) {
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__DragonFly__)
+			argv[0] = "cdcontrol";
+			argv[1] = "-f";
+			argv[2] = info->device_path?info->device_path:info->mount_point;
+			argv[3] = "eject";
+			argv[4] = NULL;
+#else
+			argv[0] = "eject";
+			argv[1] = info->device_path?info->device_path:info->mount_point;
+			argv[2] = NULL;
+#endif
+		}
+
+		error = NULL;
 		if (g_spawn_sync (NULL,
 				  argv,
 				  NULL,
@@ -328,8 +374,15 @@ mount_unmount_thread (void *arg)
 				  NULL, &standard_error,
 				  &exit_status,
 				  &error)) {
-			if (exit_status != 0 &&
-			    info->succeeded) {
+
+			if (info->succeeded == FALSE) {
+				g_free(info->error_message);
+				info->error_message = NULL;
+				g_free(info->detailed_error_message);
+				info->detailed_error_message = NULL;
+			}
+
+			if (exit_status != 0) {
 				info->succeeded = FALSE;
 				info->error_message = g_strdup (_("Unable to eject media"));
 				info->detailed_error_message = g_strdup (standard_error);
@@ -337,6 +390,8 @@ mount_unmount_thread (void *arg)
 				/* If the eject succeed then ignore the previous unmount error (if any) */
 				info->succeeded = TRUE;
 			}
+
+			g_free (standard_error);
 		} else {
 			/* Spawn failure */
 			if (info->succeeded) {
@@ -346,7 +401,6 @@ mount_unmount_thread (void *arg)
 			}
 			g_error_free (error);
 		}
-		g_free (standard_error);
 	}
 
 	g_idle_add (report_mount_result, info);	
@@ -356,63 +410,101 @@ mount_unmount_thread (void *arg)
 	return NULL;
 }
 
+#endif	/* !G_OS_WIN32 */
 
 static void
 mount_unmount_operation (const char *mount_point,
 			 const char *device_path,
+			 const char *hal_udi,
 			 GnomeVFSDeviceType device_type,
 			 gboolean should_mount,
+			 gboolean should_unmount,
 			 gboolean should_eject,
 			 GnomeVFSVolumeOpCallback  callback,
 			 gpointer                  user_data)
 {
-	const char *command;
+#ifndef G_OS_WIN32
+	const char *command = NULL;
+	const char *argument = NULL;
 	MountThreadInfo *mount_info;
 	pthread_t mount_thread;
 	const char *name;
 	int i;
 
 #ifdef USE_VOLRMMOUNT
-       name = strrchr (mount_point, '/');
-       if (name != NULL) {
-               name = name + 1;
-       } else {
-	       name = mount_point;
-       }
+	name = strrchr (mount_point, '/');
+	if (name != NULL) {
+		name = name + 1;
+	} else {
+		name = mount_point;
+	}
 #else
-       name = mount_point;
+
+# ifdef USE_HAL
+	if (hal_udi != NULL) {
+		name = device_path;
+	} else
+		name = mount_point;
+# else
+	name = mount_point;
+# endif
+
 #endif
        
        if (should_mount) {
-               command = find_command (MOUNT_COMMAND);
-       } else {
-               command = find_command (UMOUNT_COMMAND);
+#if defined(USE_HAL) && defined(HAL_MOUNT)
+	       if (hal_udi != NULL && g_file_test (HAL_MOUNT, G_FILE_TEST_IS_EXECUTABLE))
+		       command = HAL_MOUNT;
+	       else
+		       command = find_command (MOUNT_COMMAND);
+#else
+	       command = find_command (MOUNT_COMMAND);
+#endif
+#ifdef  MOUNT_ARGUMENT
+	       argument = MOUNT_ARGUMENT;
+#endif
+       }
+
+       if (should_unmount) {
+#if defined(USE_HAL) && defined(HAL_UMOUNT)
+	       if (hal_udi != NULL && g_file_test (HAL_UMOUNT, G_FILE_TEST_IS_EXECUTABLE))
+		       command = HAL_UMOUNT;
+	       else
+		       command = find_command (UMOUNT_COMMAND);
+#else
+	       command = find_command (UMOUNT_COMMAND);
+#endif
+#ifdef  UNMOUNT_ARGUMENT
+		argument = UNMOUNT_ARGUMENT;
+#endif
        }
 
 	mount_info = g_new0 (MountThreadInfo, 1);
-	i = 0;
-	mount_info->argv[i++] = g_strdup (command);
-	if (should_mount) {
-#ifdef  MOUNT_ARGUMENT
-		mount_info->argv[i++] = g_strdup (MOUNT_ARGUMENT);
-#endif
-	} else {
-#ifdef  UNMOUNT_ARGUMENT
-		mount_info->argv[i++] = g_strdup (UNMOUNT_ARGUMENT);
-#endif
+
+	if (command) {
+		i = 0;
+		mount_info->argv[i++] = g_strdup (command);
+		if (argument) {
+			mount_info->argv[i++] = g_strdup (argument);
+		}
+		mount_info->argv[i++] = g_strdup (name);
+		mount_info->argv[i++] = NULL;
 	}
-	mount_info->argv[i++] = g_strdup (name);
-	mount_info->argv[i++] = NULL;
 	
 	mount_info->mount_point = g_strdup (mount_point);
 	mount_info->device_path = g_strdup (device_path);
 	mount_info->device_type = device_type;
+	mount_info->hal_udi = g_strdup (hal_udi);
 	mount_info->should_mount = should_mount;
+	mount_info->should_unmount = should_unmount;
 	mount_info->should_eject = should_eject;
 	mount_info->callback = callback;
 	mount_info->user_data = user_data;
 	
 	pthread_create (&mount_thread, NULL, mount_unmount_thread, mount_info);
+#else
+	g_warning ("Not implemented: mount_unmount_operation()\n");
+#endif
 }
 
 
@@ -422,7 +514,7 @@ static void
 emit_pre_unmount (GnomeVFSVolume *volume)
 {
 	GnomeVFSVolumeMonitor *volume_monitor;
-#ifndef USE_EXT_DBUS_DAEMON
+#ifndef USE_DBUS_DAEMON
 	GnomeVFSClient *client;
 	GNOME_VFS_Daemon daemon;
 	CORBA_Environment  ev;
@@ -434,7 +526,7 @@ emit_pre_unmount (GnomeVFSVolume *volume)
 		gnome_vfs_volume_monitor_emit_pre_unmount (volume_monitor,
 							   volume);
 	} else {
-#ifndef USE_EXT_DBUS_DAEMON		
+#ifndef USE_DBUS_DAEMON		
 		client = _gnome_vfs_get_client ();
 		daemon = _gnome_vfs_client_get_daemon (client);
 		
@@ -551,7 +643,14 @@ gnome_vfs_volume_unmount (GnomeVFSVolume *volume,
 	char *mount_path, *device_path;
 	char *uri;
 	GnomeVFSVolumeType type;
-	
+
+	if (volume->priv->drive != NULL) {
+		if (volume->priv->drive->priv->must_eject_at_unmount) {
+			gnome_vfs_volume_eject (volume, callback, user_data);
+			return;
+		}
+	}
+
 	emit_pre_unmount (volume);
 
 	type = gnome_vfs_volume_get_volume_type (volume);
@@ -562,8 +661,9 @@ gnome_vfs_volume_unmount (GnomeVFSVolume *volume,
 		device_path = gnome_vfs_volume_get_device_path (volume);
 		mount_unmount_operation (mount_path,
 					 device_path,
+					 gnome_vfs_volume_get_hal_udi (volume),
 					 gnome_vfs_volume_get_device_type (volume),
-					 FALSE, FALSE,
+					 FALSE, TRUE, FALSE,
 					 callback, user_data);
 		g_free (mount_path);
 		g_free (device_path);
@@ -601,8 +701,9 @@ gnome_vfs_volume_eject (GnomeVFSVolume *volume,
 		device_path = gnome_vfs_volume_get_device_path (volume);
 		mount_unmount_operation (mount_path,
 					 device_path,
+					 gnome_vfs_volume_get_hal_udi (volume),
 					 gnome_vfs_volume_get_device_type (volume),
-					 FALSE, TRUE,
+					 FALSE, TRUE, TRUE,
 					 callback, user_data);
 		g_free (mount_path);
 		g_free (device_path);
@@ -634,8 +735,9 @@ gnome_vfs_drive_mount (GnomeVFSDrive  *drive,
 	device_path = gnome_vfs_drive_get_device_path (drive);
 	mount_unmount_operation (mount_path,
 				 device_path,
+				 gnome_vfs_drive_get_hal_udi (drive),
 				 GNOME_VFS_DEVICE_TYPE_UNKNOWN,
-				 TRUE, FALSE,
+				 TRUE, FALSE, FALSE,
 				 callback, user_data);
 	g_free (mount_path);
 	g_free (device_path);
@@ -659,6 +761,11 @@ gnome_vfs_drive_unmount (GnomeVFSDrive  *drive,
 	GList *vol_list;
 	GList *current_vol;
 
+	if (drive->priv->must_eject_at_unmount) {
+		gnome_vfs_drive_eject (drive, callback, user_data);
+		return;
+	}
+
 	vol_list = gnome_vfs_drive_get_mounted_volumes (drive);
 
 	for (current_vol = vol_list; current_vol != NULL; current_vol = current_vol->next) {
@@ -700,7 +807,7 @@ gnome_vfs_drive_eject (GnomeVFSDrive  *drive,
 		/* Check to see if this is the last volume */
 		/* If not simply unmount it */
 		/* If so the eject the media along with the unmount */
-		if (current_vol != NULL) { 
+		if (current_vol->next != NULL) { 
 			gnome_vfs_volume_unmount (vol,
 						  callback,
 						  user_data);
@@ -712,6 +819,23 @@ gnome_vfs_drive_eject (GnomeVFSDrive  *drive,
 
 	}
 
+	if (vol_list == NULL) { /* no mounted volumes */
+		char *mount_path, *device_path, *uri;
+	
+		uri = gnome_vfs_drive_get_activation_uri (drive);
+		mount_path = gnome_vfs_get_local_path_from_uri (uri);
+		g_free (uri);
+		device_path = gnome_vfs_drive_get_device_path (drive);
+		mount_unmount_operation (mount_path,
+					 device_path,
+					 gnome_vfs_drive_get_hal_udi (drive),
+					 GNOME_VFS_DEVICE_TYPE_UNKNOWN,
+					 FALSE, FALSE, TRUE,
+					 callback, user_data);
+		g_free (mount_path);
+		g_free (device_path);
+	}
+
 	gnome_vfs_drive_volume_list_free (vol_list);
 }
 
