@@ -64,6 +64,12 @@ enum {
 	DEFAULT_SIZE_OVERHEAD = 1024
 };
 
+/* When the total copy size is over this limit, we try to
+ * forget the data cached for the files, in order to
+ * lower VM pressure
+ */
+#define DROP_CACHE_SIZE_LIMIT (20*1024*1024)
+
 /* in asynch mode the progress callback does a context switch every time
  * it gets called. We'll only call it every now and then to not loose a
  * lot of performance
@@ -73,9 +79,9 @@ enum {
 static gint64
 system_time (void)
 {
-	struct timeval time_of_day;
+	GTimeVal time_of_day;
 
-	gettimeofday (&time_of_day, NULL);
+	g_get_current_time (&time_of_day);
 	return (gint64) time_of_day.tv_usec + ((gint64) time_of_day.tv_sec) * 1000000;
 }
 
@@ -95,6 +101,7 @@ init_progress (GnomeVFSProgressCallbackState *progress_state,
 	progress_info->bytes_copied = 0;
 	progress_info->total_bytes_copied = 0;
 	progress_info->duplicate_name = NULL;
+	progress_info->duplicate_count = 0;
 	progress_info->top_level_item = FALSE;
 
 	progress_state->progress_info = progress_info;
@@ -1201,6 +1208,7 @@ copy_file_data (GnomeVFSHandle *target_handle,
 	gpointer buffer;
 	const char *write_buffer;
 	GnomeVFSFileSize total_bytes_read;
+	gboolean forget_cache;
 
 	*skip = FALSE;
 
@@ -1210,6 +1218,9 @@ copy_file_data (GnomeVFSHandle *target_handle,
 
 	buffer = g_malloc (block_size);
 	total_bytes_read = 0;
+
+	/* Enable streaming if the total size is large */
+	forget_cache = progress->progress_info->bytes_total >= DROP_CACHE_SIZE_LIMIT;
 	
 	do {
 		GnomeVFSFileSize bytes_read;
@@ -1227,6 +1238,10 @@ copy_file_data (GnomeVFSHandle *target_handle,
 
 			result = gnome_vfs_read (source_handle, buffer,
 						 block_size, &bytes_read);
+			if (forget_cache) {
+				gnome_vfs_forget_cache (source_handle,
+							total_bytes_read, bytes_read);
+			}
 			if (result != GNOME_VFS_OK && result != GNOME_VFS_ERROR_EOF)
 				retry = handle_error (&result, progress,
 						      error_mode, skip);
@@ -1236,7 +1251,7 @@ copy_file_data (GnomeVFSHandle *target_handle,
 			break;
 		}
 
-		total_bytes_read += total_bytes_read;
+		total_bytes_read += bytes_read;
 		bytes_to_write = bytes_read;
 
 		progress->progress_info->phase = GNOME_VFS_XFER_PHASE_WRITETARGET;
@@ -1256,6 +1271,11 @@ copy_file_data (GnomeVFSHandle *target_handle,
 			bytes_to_write -= bytes_written;
 			write_buffer += bytes_written;
 		} while ((result == GNOME_VFS_OK && bytes_to_write > 0) || retry);
+		
+		if (forget_cache && bytes_to_write == 0) {
+			gnome_vfs_forget_cache (target_handle,
+						total_bytes_read - bytes_read, bytes_read);
+		}
 
 		progress->progress_info->phase = GNOME_VFS_XFER_PHASE_COPYING;
 
@@ -1503,15 +1523,17 @@ copy_file (GnomeVFSFileInfo *info,
 	if (result == GNOME_VFS_OK) {
 		/* ignore errors while setting file info attributes at this point */
 		
-		/* FIXME the modules should ignore setting of permissions if
-		 * "valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS" is clear
-		 * for now, make sure permissions aren't set to 000
-		 */
-		if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS) != 0) {
-			/* Call this separately from the time one, since one of them may fail,
-			   making the other not run. */
-			gnome_vfs_set_file_info_uri (target_uri, info, 
-						     GNOME_VFS_SET_FILE_INFO_OWNER | GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
+		if (!(xfer_options & GNOME_VFS_XFER_TARGET_DEFAULT_PERMS)) {
+			/* FIXME the modules should ignore setting of permissions if
+			 * "valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS" is clear
+			 * for now, make sure permissions aren't set to 000
+			 */
+			if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS) != 0) {
+				/* Call this separately from the time one, since one of them may fail,
+				   making the other not run. */
+				gnome_vfs_set_file_info_uri (target_uri, info, 
+							     GNOME_VFS_SET_FILE_INFO_OWNER | GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
+			}
 		}
 		
 		/* Call this last so nothing else changes the times */
@@ -1538,9 +1560,11 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 	GnomeVFSResult result;
 	GnomeVFSDirectoryHandle *source_directory_handle;
 	GnomeVFSDirectoryHandle *dest_directory_handle;
+	GnomeVFSFileInfo *target_dir_info;
 
 	source_directory_handle = NULL;
 	dest_directory_handle = NULL;
+	target_dir_info = NULL;
 	
 	result = gnome_vfs_directory_open_from_uri (&source_directory_handle, source_dir_uri, 
 						    GNOME_VFS_FILE_INFO_DEFAULT);
@@ -1574,6 +1598,14 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 		return result;
 	}
 
+	target_dir_info = gnome_vfs_file_info_new ();
+	result = gnome_vfs_get_file_info_uri (target_dir_uri, target_dir_info,
+                                        GNOME_VFS_FILE_INFO_DEFAULT);
+	if (result != GNOME_VFS_OK) {
+		gnome_vfs_file_info_unref (target_dir_info);
+		target_dir_info = NULL;
+	}
+
 	if (call_progress_with_uris_often (progress, source_dir_uri, target_dir_uri, 
 					   GNOME_VFS_XFER_PHASE_OPENTARGET) != 0) {
 
@@ -1600,6 +1632,9 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 				gnome_vfs_file_info_unref (info);	
 				break;
 			}
+			if (target_dir_info && GNOME_VFS_FILE_INFO_SGID(target_dir_info)) {
+				info->gid = target_dir_info->gid;
+			}
 			
 			/* Skip "." and "..".  */
 			if (strcmp (info->name, ".") != 0 && strcmp (info->name, "..") != 0) {
@@ -1608,7 +1643,8 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 
 				source_uri = gnome_vfs_uri_append_file_name (source_dir_uri, info->name);
 				dest_uri = gnome_vfs_uri_append_file_name (target_dir_uri, info->name);
-				
+				progress_set_source_target_uris (progress, source_uri, dest_uri);
+
 				if (info->type == GNOME_VFS_FILE_TYPE_REGULAR) {
 					result = copy_file (info, source_uri, dest_uri, 
 							    xfer_options, error_mode, overwrite_mode, 
@@ -1660,21 +1696,40 @@ copy_directory (GnomeVFSFileInfo *source_file_info,
 	gnome_vfs_directory_close (dest_directory_handle);
 	gnome_vfs_directory_close (source_directory_handle);
 
-	if (result == GNOME_VFS_OK) {
-		/* FIXME the modules should ignore setting of permissions if
-		 * "valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS" is clear
-		 * for now, make sure permissions aren't set to 000
-		 */
+	if (result == GNOME_VFS_OK) {			
+		GnomeVFSFileInfo *info;
 
-		if ((source_file_info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS) != 0) {
-			/* Call this separately from the time one, since one of them may fail,
-			   making the other not run. */
-			gnome_vfs_set_file_info_uri (target_dir_uri, source_file_info, 
-						     GNOME_VFS_SET_FILE_INFO_OWNER | GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
+		if (target_dir_info && GNOME_VFS_FILE_INFO_SGID (target_dir_info)) {
+			info = gnome_vfs_file_info_dup (source_file_info);
+			info->gid = target_dir_info->gid;
+		} else {
+			/* No need to dup the file info in this case */
+			gnome_vfs_file_info_ref (source_file_info);
+			info = source_file_info;
+		}
+
+		if (!(xfer_options & GNOME_VFS_XFER_TARGET_DEFAULT_PERMS)) {
+			/* FIXME the modules should ignore setting of permissions if
+			 * "valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS" is clear
+			 * for now, make sure permissions aren't set to 000
+			 */
+			if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS) != 0) {
+				
+				/* Call this separately from the time one, since one of them may fail,
+				   making the other not run. */
+				gnome_vfs_set_file_info_uri (target_dir_uri, info, 
+							     GNOME_VFS_SET_FILE_INFO_OWNER | 
+							     GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
+			}
 		}
 		
 		/* Call this last so nothing else changes the times */
-		gnome_vfs_set_file_info_uri (target_dir_uri, source_file_info, GNOME_VFS_SET_FILE_INFO_TIME);
+		gnome_vfs_set_file_info_uri (target_dir_uri, info, GNOME_VFS_SET_FILE_INFO_TIME);
+		gnome_vfs_file_info_unref (info);
+	}
+	
+	if (target_dir_info) {
+		gnome_vfs_file_info_unref (target_dir_info);
 	}
 
 	return result;
@@ -1725,6 +1780,17 @@ copy_items (const GList *source_uri_list,
 			((GnomeVFSURI *)target_item->data);
 
 		if (result == GNOME_VFS_OK) {
+			GnomeVFSFileInfo *target_dir_info;
+			/* get target_dir URI and file info to take care of SGID */
+			target_dir_info = gnome_vfs_file_info_new ();
+			result = gnome_vfs_get_file_info_uri (target_dir_uri, target_dir_info,
+																						GNOME_VFS_FILE_INFO_DEFAULT);
+			if (result == GNOME_VFS_OK && GNOME_VFS_FILE_INFO_SGID(target_dir_info)) {
+				info->gid = target_dir_info->gid;
+			}
+			gnome_vfs_file_info_unref (target_dir_info);
+			result = GNOME_VFS_OK;
+
 			/* optionally keep trying until we hit a unique target name */
 			for (count = 1; ; count++) {
 				GnomeVFSXferOverwriteMode overwrite_mode_abort;
@@ -2125,11 +2191,11 @@ gnome_vfs_xfer_delete_items_common (const GList *source_uri_list,
 					      &skip);
 		}
 
-		gnome_vfs_file_info_unref (info);
-
 		if (result != GNOME_VFS_OK) {
 			break;
 		}
+
+		gnome_vfs_file_info_unref (info);
 	}
 
 	return result;
@@ -2236,38 +2302,34 @@ gnome_vfs_new_directory_with_unique_name (const GnomeVFSURI *target_dir_uri,
 }
 
 static GnomeVFSResult
-gnome_vfs_destination_is_writable (const GnomeVFSURI *uri)
+gnome_vfs_destination_is_writable (GnomeVFSURI *uri)
 {
-	GnomeVFSURI *test_uri;
 	GnomeVFSResult result;
-	GnomeVFSHandle *handle;
-
-	if (!gnome_vfs_uri_is_local (uri)) {
-		/* if destination is not local, do not test it for writability, just
-		 * assume it's writable
-		 */
-		return GNOME_VFS_OK;
-	}
-
-	/* test writability by creating and erasing a temporary file */
-	test_uri = gnome_vfs_uri_append_file_name (uri, ".vfs-write.tmp");
-	result = gnome_vfs_create_uri (&handle, test_uri, GNOME_VFS_OPEN_WRITE, TRUE, 0600);
-
-	if (result == GNOME_VFS_OK) {
-		gnome_vfs_close (handle);
-	}
+	GnomeVFSFileInfo *info;
 	
-	if (result == GNOME_VFS_OK || result == GNOME_VFS_ERROR_FILE_EXISTS) {
-		gnome_vfs_unlink_from_uri (test_uri);
-		result = GNOME_VFS_OK;
-	}
+	info   = gnome_vfs_file_info_new ();	    
+	result = gnome_vfs_get_file_info_uri (uri, info,
+					      GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS);
 	
-	/* some methods only allow certain filenames (e.g. .desktop files) */
-	if (result == GNOME_VFS_ERROR_INVALID_URI) {
-		result = GNOME_VFS_OK;
+		
+	if (result != GNOME_VFS_OK) {
+		gnome_vfs_file_info_unref (info);
+		return result;
 	}
 
-	gnome_vfs_uri_unref (test_uri);
+	/* Default to GNOME_VFS_OK so we return that if we don't have support
+	   for GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS */
+	result = GNOME_VFS_OK;
+	
+	if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_ACCESS) {
+		
+		if (! (info->permissions & GNOME_VFS_PERM_ACCESS_WRITABLE &&
+		       info->permissions & GNOME_VFS_PERM_ACCESS_EXECUTABLE)) {
+			result = GNOME_VFS_ERROR_ACCESS_DENIED;
+		}
+	}
+
+	gnome_vfs_file_info_unref (info);
 	return result;
 }
 
@@ -2686,6 +2748,9 @@ gnome_vfs_xfer_uri_list (const GList *source_uri_list,
  * onto the destination is not overwritten.  It will only copy the unique items
  * from the source to the destjnation.
  * GNOME_VFS_XFER_LINK_ITEMS: <TBA>
+ * GNOME_VFS_XFER_TARGET_DEFAULT_PERMS: This means that the target file will
+ * not have the same permissions of the source file, but will instead have the 
+ * default permissions of the destination location.
  * @error_mode:  When this function returns you need to check the error code
  * for the results of the copy.  The results are generally:
  * GNOME_VFS_XFER_ERROR_MODE_ABORT: This means that the operation was aborted
